use std::{
    collections::HashMap,
    env::current_dir,
    fs,
    io::{BufRead, Write},
    path::PathBuf,
};

use clap::Parser;

use crate::{
    config::read_config,
    module::{direnv::DirenvModule, gitignore::GitIgnoreModule, nix_flake::NixFlakeModule, Module},
};

use self::{
    direnv::DirenvModuleCliArgs, gitignore::GitIgnoreModuleCliArgs,
    nix_flake::NixFlakeModuleCliArgs,
};

mod direnv;
mod gitignore;
mod nix_flake;

#[derive(Debug, Clone, clap::Args)]
struct MakeCommand {
    modules: Option<Vec<String>>,
}

#[derive(Debug, Clone, clap::Subcommand)]
enum AddModuleCommand {
    Direnv(DirenvModuleCliArgs),
    #[clap(name = "gitignore")]
    GitIgnore(GitIgnoreModuleCliArgs),
    #[clap(name = "nix.flake")]
    NixFlake(NixFlakeModuleCliArgs),
    #[clap(external_subcommand)]
    External(Vec<String>),
}

#[derive(Debug, Clone, clap::Args)]
struct AddCommand {
    #[clap(trailing_var_arg = true, allow_hyphen_values = true)]
    module_args: Option<Vec<String>>,
}

#[derive(Debug, Clone, clap::Subcommand)]
#[clap(propagate_version = true)]
enum Command {
    Make(MakeCommand),
    Add {
        #[clap(short, long)]
        force: bool,
        #[clap(subcommand)]
        command: AddModuleCommand,
    },
}

#[derive(Debug, clap::Parser)]
#[clap(author, version)]
struct Args {
    #[clap(short, long = "path")]
    project_path: Option<PathBuf>,
    #[clap(subcommand)]
    command: Command,
}

fn tui_question(question: &str, default: bool) -> bool {
    print!("{question}: ({}) ", if default { "Y/n" } else { "y/N" });

    std::io::stdout()
        .flush()
        .expect("Flushing stdout shouldn't fail");
    let stdin = std::io::stdin();
    let mut handle = stdin.lock();
    let value = loop {
        let mut buffer = String::new();
        handle
            .read_line(&mut buffer)
            .expect("Input string shouldn't fail");
        break match buffer.to_lowercase().trim() {
            "" => default,
            "y" => true,
            "n" => false,
            _ => continue,
        };
    };

    value
}

fn make(
    current_dir: PathBuf,
    make_files: HashMap<String, String>,
    force: bool,
) -> Result<(), Box<dyn std::error::Error>> {
    for (file_name, file_content) in make_files {
        let file_path = &current_dir.join(file_name);
        let file_exists = file_path.exists();
        let is_empty_file_content = file_content.trim().is_empty();

        println!("{}", file_path.to_str().unwrap());
        if !force && file_exists {
            if !(tui_question(
                &format!(
                    "    > already exists. Are you sure you want to {} it?",
                    if is_empty_file_content {
                        "delete"
                    } else {
                        "overwrite"
                    }
                ),
                false,
            )) {
                println!("    > skipped");
                continue;
            }
        }

        if is_empty_file_content {
            if file_exists {
                fs::remove_file(file_path)?;
                println!("    > deleted")
            }
        } else {
            fs::write(file_path, file_content)?;
            if file_exists {
                println!("    > overwritten")
            } else {
                println!("    > created")
            }
        }
    }

    Ok(())
}

pub fn run() -> Result<(), Box<dyn std::error::Error>> {
    let args = Args::parse();
    let current_dir = if let Some(project_path) = args.project_path {
        project_path.canonicalize()?
    } else {
        current_dir()?
    };

    match args.command {
        Command::Make(_make_args) => {
            let config = read_config(&current_dir)?;
            let make_files: HashMap<String, String> = [
                config.direnv.map(|v| DirenvModule.make(v.into())),
                config.gitignore.map(|v| GitIgnoreModule.make(v.into())),
                config
                    .nix
                    .and_then(|cfg| cfg.flake.map(|v| NixFlakeModule.make(v.into()))),
            ]
            .into_iter()
            .flatten()
            .fold(HashMap::new(), |mut acc, files| {
                acc.extend(files);
                acc
            });

            make(current_dir, make_files, true)
        }
        Command::Add { force, command } => {
            let make_files: HashMap<String, String> = match command {
                AddModuleCommand::Direnv(args) => DirenvModule.make(args.into()),
                AddModuleCommand::GitIgnore(args) => GitIgnoreModule.make(args.into()),
                AddModuleCommand::NixFlake(args) => NixFlakeModule.make(args.into()),
                AddModuleCommand::External(_) => {
                    unimplemented!("external commands")
                }
            };

            make(current_dir, make_files, force)
        }
    }
}
