use std::{fs, io, path::PathBuf};

use self::{
    direnv::DirenvModuleConfig, gitignore::GitIgnoreModuleConfig, nix_flake::NixFlakeModuleConfig,
};

mod direnv;
mod gitignore;
mod nix_flake;

#[derive(serde::Deserialize, serde::Serialize)]
pub struct NixConfig {
    pub flake: Option<NixFlakeModuleConfig>,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct Config {
    pub nix: Option<NixConfig>,
    pub direnv: Option<DirenvModuleConfig>,
    pub gitignore: Option<GitIgnoreModuleConfig>,
}

#[derive(Debug)]
pub enum Error {
    ReadFile,
    Deserialize,
    Unknown,
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::ReadFile => f.write_str("Cannot read file"),
            Self::Deserialize => f.write_str("Cannot deserialize configuration"),
            Self::Unknown => f.write_str("Unknown error"),
        }
    }
}

impl std::error::Error for Error {}

impl From<io::Error> for Error {
    fn from(value: io::Error) -> Self {
        match value.kind() {
            io::ErrorKind::NotFound | io::ErrorKind::InvalidData => Self::ReadFile,
            _ => Self::Unknown,
        }
    }
}

impl From<toml::de::Error> for Error {
    fn from(_value: toml::de::Error) -> Self {
        Self::Deserialize
    }
}

pub fn read_config(workdir: &PathBuf) -> Result<Config, Error> {
    let content = fs::read_to_string(workdir.join("fimt.toml"))?;
    let res = toml::from_str(&content).inspect_err(|e| {
        dbg!(e);
    })?;
    Ok(res)
}
